@@ -54,6 +54,7 @@
 */
 
 #include "dwarfvm.h"
+#include "leb.h"
 #include "register.h"
 #include <dwarf.h>
 #include "util/dictionary.h"
@@ -63,6 +64,8 @@
 #include "symbol.h"
 #include "util/map.h"
 #include "patcher/hotpatch.h"
+#include "util/stack.h"
+#include "elfutil.h"
 
 //returns a list of PatchData objects
 List* generatePatchesFromFDEAndState(FDE* fde,SpecialRegsState* state,ElfInfo* patch,ElfInfo* patchedBin);
@@ -70,59 +73,137 @@ List* generatePatchesFromFDEAndState(FDE* fde,SpecialRegsState* state,ElfInfo* p
 //keys are old addresses of data objects (variables). Values are the new addresses
 Map* dataMoved=NULL;
 
-//evaluates the given instructions and stores them in the output rules dictionary
-//the initial condition of regarray IS taken into account
-//execution continues until the end of the instructions or until the location is advanced
-//past stopLocation. stopLocation should be relative to the start of the instructions (i.e. the instructions are considered to start at 0)
-//if stopLocation is negative, it is ignored
-void evaluateInstructionsToRules(RegInstruction* instrs,int numInstrs,Dictionary* rules,int stopLocation)
+//this is the stack of saved register states used by the
+//DW_CFA_remember_state and DW_CFA_restore_state instructions
+static Stack* stateStack;
+//unfortunately because the way we currently evaluate
+//static Map* rememberedStatesByLocation;
+
+//evaluates the given instructions and stores them in the output rules
+//dictionary.The initial condition of regarray IS taken into account.
+//Execution continues until the end of the instructions or until the
+//location is advanced past stopLocation. stopLocation should be
+//relative to the start of the instructions (i.e. the instructions are
+//considered to start at 0) if stopLocation is negative, it is ignored
+//if stopLocation is negative, it is ignored (evaluation continues to
+//the end of the instructions) returns the location stopped at (will
+//be the lowest location that a change was actually made).
+//outInstrsCnt, if non-NULL, is used to store the number of instructions read
+int evaluateInstructionsToRules(CIE* cie,RegInstruction* instrs,int numInstrs,Dictionary* rules,int startLocation, int stopLocation,int* outInstrsCnt)
 {
   PoRegRule* rule=NULL;
-  int loc=0;
+  int loc=startLocation;
   for(int i=0;i<numInstrs;i++)
   {
     RegInstruction inst=instrs[i];
-    if(ERT_NONE==inst.arg1Reg.type)
+
+    //deal first with the location-advancing instructions and other
+    //instructions (remember/restore state) that don't do the ordinary
+    //register manipulation
+    switch(inst.type)
     {
-      death("cannot evaluate register with type ERT_NONE\n");
+    case DW_CFA_set_loc:
+      if(stopLocation >= 0 && inst.arg1>stopLocation)
+      {
+        return loc;
+      }
+      loc=inst.arg1;
+      continue;
+    case DW_CFA_advance_loc:
+    case DW_CFA_advance_loc1:
+    case DW_CFA_advance_loc2:
+      if(stopLocation >=0 && loc+inst.arg1>stopLocation)
+      {
+        return loc;
+      }
+      loc+=inst.arg1*cie->codeAlign;
+      continue;
+    case DW_CFA_remember_state:
+      {
+        //Note that the storage used for remember and restore state is
+        //global.  no attempt is made to verify that a
+        //DW_CFA_restore_state is used in a sensible manner. It will
+        //restore whatever was last pushed onto the stack when Katana is
+        //running. It is assumed that the DWARF file will be constructed
+        //in a sensible manner. Otherwise the generated rules may make
+        //little sense.
+        Dictionary* rulesCopy=dictDuplicate(rules,(DictDataCopy)duplicatePoRegRule);
+        if(!stateStack)
+        {
+          stateStack=stackCreate();
+        }
+        stackPush(stateStack,rulesCopy);
+        continue;
+      }
       break;
+    case DW_CFA_restore_state:
+      {
+        if(!stateStack)
+        {
+          death("Attempt to use DW_CFA_restore_state without using DW_CFA_remember_state\n");
+        }
+        Dictionary* savedRules=stackPop(stateStack);
+        if(!savedRules)
+        {
+          death("Attempt to use DW_CFA_restore_state without using DW_CFA_remember_state\n");
+        }
+        char** keys=dictKeys(savedRules);
+        for(int i=0;keys[i];i++)
+        {
+          char* key=keys[i];
+          PoRegRule* currentRule=dictGet(rules,key);
+          PoRegRule* savedRule=dictGet(savedRules,key);
+          assert(savedRule);
+          if(currentRule)
+          {
+            memcpy(currentRule,savedRule,sizeof(PoRegRule));
+          }
+          else
+          {
+            currentRule=duplicatePoRegRule(savedRule);
+            dictInsert(rules,key,currentRule);
+          }
+        }
+        dictDelete(savedRules,free);
+        continue;
+      }
+    }
+
+    
+    
+    PoReg reg;
+    char* str=NULL;
+    if(DW_CFA_def_cfa==inst.type ||
+       DW_CFA_def_cfa_register==inst.type ||
+       DW_CFA_def_cfa_offset==inst.type)
+    {
+      memset(&reg,0,sizeof(reg));
+      reg.type=ERT_CFA;
     }
     else
     {
-      PoReg reg;
-      char* str=NULL;
-      if(DW_CFA_def_cfa==inst.type ||
-         DW_CFA_def_cfa_register==inst.type ||
-         DW_CFA_def_cfa_offset==inst.type)
-      {
-        memset(&reg,0,sizeof(reg));
-        reg.type=ERT_CFA;
-      }
-      else
-      {
-        reg=inst.arg1Reg;
-      }
-      str=strForReg(reg);
-              
-      rule=dictGet(rules,str);
-      if(!rule)
-      {
-        rule=zmalloc(sizeof(PoRegRule));
-        rule->regLH=reg;
-        dictInsert(rules,str,rule);
-        
-      }
-      free(str);
+      reg=inst.arg1Reg;
     }
+    str=strForReg(reg,0);
+              
+    rule=dictGet(rules,str);
+    if(!rule)
+    {
+      rule=zmalloc(sizeof(PoRegRule));
+      rule->regLH=reg;
+      dictInsert(rules,str,rule);
+        
+    }
+    free(str);
     //printf("evaluating instruction of type 0x%x\n",(uint)inst.type);
     switch(inst.type)
     {
     case DW_CFA_set_loc:
-      loc=inst.arg1;
-      if(stopLocation >= 0 && loc>stopLocation)
+      if(stopLocation >= 0 && loc+inst.arg1>stopLocation)
       {
-        return;
+        return loc;
       }
+      loc=inst.arg1;
       break;
     case DW_CFA_advance_loc:
     case DW_CFA_advance_loc1:
@@ -130,12 +211,12 @@ void evaluateInstructionsToRules(RegInstruction* instrs,int numInstrs,Dictionary
       loc+=inst.arg1;
       if(stopLocation >= 0 && loc>stopLocation)
       {
-        return;
+        return loc;
       }
       break;
     case DW_CFA_offset:
       rule->type=ERRT_OFFSET;
-      rule->offset=inst.arg2;
+      rule->offset=inst.arg2*cie->dataAlign;
       break;
     case DW_CFA_register:
       rule->type=ERRT_REGISTER;
@@ -154,6 +235,20 @@ void evaluateInstructionsToRules(RegInstruction* instrs,int numInstrs,Dictionary
       rule->type=ERRT_CFA;
       rule->offset=inst.arg1;
       break;
+    case DW_CFA_restore:
+      {
+        PoRegRule* initialRule=dictGet(cie->initialRules,str);
+        if(initialRule)
+        {
+          *rule=*initialRule;
+        }
+        else
+        {
+          //there was no inital rule. Is there ever a valid usage for this?
+          death("Unable to create DW_CFA_restore rule because no rule assigned in the CIE\n");
+        }
+      }
+      break;
     case DW_CFA_KATANA_fixups:
       rule->type=ERRT_RECURSE_FIXUP;
       rule->regRH=inst.arg2Reg;
@@ -164,13 +259,19 @@ void evaluateInstructionsToRules(RegInstruction* instrs,int numInstrs,Dictionary
       rule->regRH=inst.arg2Reg;
       rule->index=inst.arg3;
       break;
+    case DW_CFA_expression:
+    case DW_CFA_val_expression:
+      //todo: support this
+      logprintf(ELL_WARN,ELS_VM,"DW_CFA_expression/DW_CFA_val_expression is not supported in the DWARF VM yet. Ignoring this instruction. This means that the register rules will not be accurate.\n");
+      break;
     case DW_CFA_nop:
       //do nothing, nothing changed
       break;
     default:
-      death("unexpected instruction in evaluateInstructions");
+      death("unexpected instruction in dwarfvm evaluateInstructions");
     }
   }
+  return loc;
 }
 
 //structure for holding one contiguous chunk of data
@@ -261,7 +362,7 @@ List* makePatchData(PoRegRule* rule,SpecialRegsState* state,ElfInfo* patch,ElfIn
       memcpy(&tmpState.currAddrOld,rhAddrBytes,sizeof(addr_t));
       free(rhAddrBytes);
       //fde indices seem to be 1-based and we store them zero-based
-      head=generatePatchesFromFDEAndState(&patch->fdes[rule->index-1],&tmpState,patch,patchedBin);
+      head=generatePatchesFromFDEAndState(&patch->callFrameInfo.fdes[rule->index-1],&tmpState,patch,patchedBin);
     }
     break;
   case ERRT_RECURSE_FIXUP_POINTER:
@@ -353,7 +454,7 @@ List* makePatchData(PoRegRule* rule,SpecialRegsState* state,ElfInfo* patch,ElfIn
         //      lacking important information)
         
         //pointedObjectNewLocation=getFreeSpaceInTarget(patch->fdes[rule->index-1].memSize);
-        pointedObjectNewLocation=mallocTarget(patch->fdes[rule->index-1].memSize);
+        pointedObjectNewLocation=mallocTarget(patch->callFrameInfo.fdes[rule->index-1].memSize);
         logprintf(ELL_INFO_V2,ELS_DWARF_FRAME,"No symbol associated with object at address 0x%zx we have to relocate that we have a pointer to. Mallocced new memory at 0x%zx\n",tmpState.currAddrOld,pointedObjectNewLocation);
       }
 
@@ -368,7 +469,7 @@ List* makePatchData(PoRegRule* rule,SpecialRegsState* state,ElfInfo* patch,ElfIn
       memcpy(result->data,&pointedObjectNewLocation,sizeof(addr_t));
       
       //fde indices seem to be 1-based and we store them zero-based
-      head->next=generatePatchesFromFDEAndState(&patch->fdes[rule->index-1],&tmpState,patch,patchedBin);
+      head->next=generatePatchesFromFDEAndState(&patch->callFrameInfo.fdes[rule->index-1],&tmpState,patch,patchedBin);
     }
     break;
   default:
@@ -394,7 +495,7 @@ List* generatePatchesFromFDEAndState(FDE* fde,SpecialRegsState* state,ElfInfo* p
   //we build up rules for each register from the DW_CFA instructions
   Dictionary* rulesDict=dictCreate(100);//todo: get rid of arbitrary constant 100
   //todo: versioning?
-  evaluateInstructionsToRules(fde->instructions,fde->numInstructions,rulesDict,fde->highpc);
+  evaluateInstructionsToRules(fde->cie,fde->instructions,fde->numInstructions,rulesDict,fde->lowpc,fde->highpc,NULL);
   PoRegRule** rules=(PoRegRule**)dictValues(rulesDict);
   //we gather all of the the patch data together first before actually poking the target
   //because everything is supposed to be applied in parallel, as a table, and
@@ -406,7 +507,7 @@ List* generatePatchesFromFDEAndState(FDE* fde,SpecialRegsState* state,ElfInfo* p
   PoReg cfaReg;
   memset(&cfaReg,0,sizeof(PoReg));
   cfaReg.type=ERT_CFA;
-  char* str=strForReg(cfaReg);
+  char* str=strForReg(cfaReg,0);
   PoRegRule* cfaRule=dictGet(rulesDict,str);
   free(str);
   if(cfaRule)
