@@ -52,8 +52,8 @@
   Date: January 2010
   Description: Read patch object files and apply them
 */
-
-#include "elfparse.h"
+
+#include "elfutil.h"
 #include "dwarftypes.h"
 #include "hotpatch.h"
 #include "target.h"
@@ -75,6 +75,7 @@
 #include <unistd.h>
 #include "pmap.h"
 #include "patchapply.h"
+#include "config.h"
 
 ElfInfo* patchedBin=NULL;
 ElfInfo* targetBin=NULL;
@@ -357,7 +358,7 @@ addr_t copyInEntireSection(ElfInfo* patch,char* name,char* newName)
   shdrNew.sh_addr=addr;
   shdrNew.sh_name=addStrtabEntryToExisting(patchedBin,newName,true);
   shdrNew.sh_type=shdr.sh_type;
-  shdrNew.sh_flags=shdr.sh_flags;
+  shdrNew.sh_flags=shdr.sh_flags | SHF_ALLOC;
   shdrNew.sh_size=shdr.sh_size;
   shdrNew.sh_link=0;//todo: should this be set?
   shdrNew.sh_info=0;//todo: should this be set?
@@ -382,10 +383,11 @@ addr_t copyInEntireSection(ElfInfo* patch,char* name,char* newName)
 }
 
 
-//this is a horrible function full of hacks that I've been trying to get
-//working and have been failing it. It should be massively revamped before it's any good
-//basically it attempts to fix the program header so the program can run, but doesn't
-//do so successfully
+//this is a horrible function full of hacks that I've been trying to
+//get working and have been failing it. It should be massively
+//revamped before it's any good basically it attempts to fix the
+//program header so the program can run, but doesn't do so
+//successfully
 void writeOutPatchedBin(bool flushToDisk)
 {
   //todo: this is all not working
@@ -567,18 +569,17 @@ void fixupPatchRelocations(ElfInfo* patch)
     //if some not taken care of?
     if(type==R_386_PC32 || type==R_X86_64_PC32)
     {
-      //todo: BROKEN. Shouldn't be assuming we're using textData, might not be in text
-        assert(rela->r_offset<textData->d_size);
-        word_t addrAccessed=*((word_t*)(textData->d_buf+rela->r_offset));
-        //if we were using a PC-relative relocation and the PC wasn't in the patch text,
-        //have to update it for this relocation
-        //todo: don't do this if the relocation was actually
-        //for something relative to the patch
-        addr_t diff=patchTextAddr-oldTextNewStart;
-        logprintf(ELL_INFO_V2,ELS_RELOCATION,"for PC32 relocation, modifying access at 0x%x to access 0x%x by adding 0x%x\n",(uint)newOffset,(uint)(addrAccessed+diff),(uint)diff);
-        addr_t newAddr=addrAccessed+diff;
+      assert(rela->r_offset+sizeof(uint32) <= textData->d_size);
+      addr_t addrAccessed=(addr_t)*((uint32*)(textData->d_buf+rela->r_offset));
+      //if we were using a PC-relative relocation and the PC wasn't in the patch text,
+      //have to update it for this relocation
+      //todo: don't do this if the relocation was actually
+      //for something relative to the patch
+      addr_t diff=patchTextAddr-oldTextNewStart;
+      logprintf(ELL_INFO_V2,ELS_RELOCATION,"for PC32 relocation, modifying access at 0x%x to access 0x%x by adding 0x%x\n",(uint)newOffset,(uint)(addrAccessed+diff),(uint)diff);
+      addr_t newAddr=addrAccessed+diff;
         
-        memcpyToTarget(newOffset,(byte*)&newAddr,4);//always copy only 4 bytes b/c this is PC32
+      memcpyToTarget(newOffset,(byte*)&newAddr,4);//always copy only 4 bytes b/c this is PC32
         //todo: I'm not sure this is necessary here, since we do
         //relocations later. I think it is though. Look into this
     }
@@ -590,6 +591,7 @@ void fixupPatchRelocations(ElfInfo* patch)
   }
 }
 
+//copy PLT and GOT to new locations, as we may need to expand them
 ///todo: this function may not work for x86-64 Large Code Model
 void katanaPLT()
 {
@@ -766,7 +768,7 @@ void katanaPLT()
 
 }
 
-void readAndApplyPatch(int pid,ElfInfo* targetBin_,ElfInfo* patch,int flags)
+void readAndApplyPatch(int pid,ElfInfo* targetBin_,ElfInfo* patch)
 {
   startPtrace(pid);
   targetBin=targetBin_;
@@ -805,7 +807,7 @@ void readAndApplyPatch(int pid,ElfInfo* targetBin_,ElfInfo* patch,int flags)
   char cwd[PATH_MAX];
   getcwd(cwd,PATH_MAX);
   DwarfInfo* diPatch=readDWARFTypes(patch,cwd);
-  Map* fdeMap=readDebugFrame(patch);//get mapping between fde offsets and fde structures
+  Map* fdeMap=readDebugFrame(patch,false);//get mapping between fde offsets and fde structures
 
 
   //we need to know where malloc lives in the target because
@@ -841,8 +843,6 @@ void readAndApplyPatch(int pid,ElfInfo* targetBin_,ElfInfo* patch,int flags)
   getShdrByERS(targetBin,ERS_GOTPLT,&shdr);
   amount+=shdr.sh_size;
 
-
-
   #ifdef KATANA_X86_64_ARCH
   addr_t desiredAddress=0;
 
@@ -860,9 +860,6 @@ void readAndApplyPatch(int pid,ElfInfo* targetBin_,ElfInfo* patch,int flags)
   }
   #endif
 
-  
-
-
   #ifdef KATANA_X86_64_ARCH
   addr_t receivedAddres=reserveFreeSpaceInTarget(amount,desiredAddress);
   if(receivedAddres > 0xFFFFFFFF && patchedBin->textUsesSmallCodeModel)
@@ -948,8 +945,7 @@ void readAndApplyPatch(int pid,ElfInfo* targetBin_,ElfInfo* patch,int flags)
   writeOutPatchedBin(true);
   endELF(targetBin);
   endELF(patchedBin);
-  endELF(patch);
   cleanupDwarfVM();
-  endPtrace(flags & PF_STOP_TARGET);
+  endPtrace(isFlag(EKCF_P_STOP_TARGET));
   printf("hooray! completed application of patch successfully\n");
 }
