@@ -52,7 +52,7 @@
   Date: January 2010
   Description: Write patch information out to a po (patch object) file
 */
-
+
 #include "types.h"
 #include "dwarftypes.h"
 #include <gelf.h>
@@ -70,6 +70,7 @@
 #include "elfwriter.h"
 #include "sourcetree.h"
 #include "write_to_dwarf.h"
+#include "elfutil.h"
 
 ElfInfo* oldBinary=NULL;
 ElfInfo* newBinary=NULL;
@@ -77,7 +78,6 @@ ElfInfo* newBinary=NULL;
 
 ElfInfo* patch;
 addr_t rodataOffset;//used to keep track of how much rodata we've used
-Elf *outelf;
 
 
 Dwarf_P_Die firstCUDie=NULL;
@@ -109,7 +109,7 @@ idx_t addSymbolFromBinaryToPatch(ElfInfo* binary,idx_t symIdx)
   //that this relocation refers to. If we don't already have a corresponding
   //symbol in the patch file, we may have to add this one
   ElfXX_Sym sym;
-  sym.st_name=addStrtabEntry(symname);//todo: this is a waste if we don't end up using this symbol
+  sym.st_name=addStrtabEntry(patch,symname);//todo: this is a waste if we don't end up using this symbol
   sym.st_value=symInNewBin.st_value;
   sym.st_size=symInNewBin.st_size;
   sym.st_info=symInNewBin.st_info;
@@ -117,7 +117,7 @@ idx_t addSymbolFromBinaryToPatch(ElfInfo* binary,idx_t symIdx)
   sym.st_shndx=SHN_UNDEF;        
   if(STT_SECTION==symType)
   {
-    sym.st_shndx=reindexSectionForPatch(newBinary,symInNewBin.st_shndx);
+    sym.st_shndx=reindexSectionForPatch(newBinary,symInNewBin.st_shndx,patch);
   }
   sym.st_info=ELFXX_ST_INFO(ELF64_ST_BIND(symInNewBin.st_info),
                             symType);
@@ -128,7 +128,7 @@ idx_t addSymbolFromBinaryToPatch(ElfInfo* binary,idx_t symIdx)
   int reindex=findSymbol(patch,&gsym,patch,false);
   if(SHN_UNDEF==reindex)
   {
-    reindex=addSymtabEntry(getDataByERS(patch,ERS_SYMTAB),&sym);
+    reindex=addSymtabEntry(patch,getDataByERS(patch,ERS_SYMTAB),&sym);
   }
   return reindex;
 }
@@ -268,10 +268,10 @@ void writeVarToData(VarInfo* var)
   }
   //now we must create a symbol for it as well
   ElfXX_Sym symNew=gelfSymToNativeSym(sym);
-  symNew.st_name=addStrtabEntry(var->name);
+  symNew.st_name=addStrtabEntry(patch,var->name);
   symNew.st_value=addr;//will get relocated later once the data section in the patch has a fixed address
   symNew.st_shndx=elf_ndxscn(getSectionByERS(patch,ERS_DATA));
-  addSymtabEntry(getDataByERS(patch,ERS_SYMTAB),&symNew);
+  addSymtabEntry(patch,getDataByERS(patch,ERS_SYMTAB),&symNew);
 }
 
 //takes a list of var transformations and actually constructs the ELF sections
@@ -507,14 +507,14 @@ addr_t writeFuncToPatchText(SubprogramInfo* func,CompilationUnit* cu,idx_t* outS
   addr_t funcSegmentBase=offset+shdr.sh_addr;//where text for this function is based
   //if -ffunction-sections is used, the function might have its own text section
   ElfXX_Sym sym;
-  sym.st_name=addStrtabEntry(func->name);
+  sym.st_name=addStrtabEntry(patch,func->name);
   sym.st_value=offset;//is it ok that this is section-relative, since
   //we set st_shndx to be the text section?
   sym.st_size=0;
   sym.st_info=ELFXX_ST_INFO(STB_GLOBAL,STT_FUNC);
   sym.st_other=0;
   sym.st_shndx=elf_ndxscn(textScn);
-  *outSymIdx=addSymtabEntry(getDataByERS(patch,ERS_SYMTAB),&sym);
+  *outSymIdx=addSymtabEntry(patch,getDataByERS(patch,ERS_SYMTAB),&sym);
   
 
   Elf_Scn* relocScn=getRelocationSection(cu->elf,func->name);
@@ -739,13 +739,21 @@ void writeROData(ElfInfo* binary)
 }
 
 
-void writePatch(char* oldSourceTree,char* newSourceTree,char* oldBinName,char* newBinName,char* patchOutName)
+ElfInfo* createPatch(char* oldSourceTree,char* newSourceTree,char* oldBinName,char* newBinName,FILE* patchOutfile,char* filename)
 {
+  if(!patchOutfile)
+  {
+    patchOutfile=fopen(filename,"w");
+    if(!patchOutfile)
+    {
+      death("Could not create patch because could not open file %s for writing\n",filename);
+    }
+  }
   newBinary=openELFFile(newBinName);
   findELFSections(newBinary);
   oldBinary=openELFFile(oldBinName);
   findELFSections(oldBinary);
-  patch=startPatchElf(patchOutName);
+  patch=startPatchElf(patchOutfile,filename);
   Dwarf_Error err;
   //todo: always creating 32-bit Dwarf sections. DWARF standard
   //recommends never having 64-bit Dwarf as the default, even on
@@ -825,7 +833,7 @@ void writePatch(char* oldSourceTree,char* newSourceTree,char* oldBinName,char* n
     Dwarf_Signed elfScnIdx;
     Dwarf_Unsigned length;
     Dwarf_Ptr buf=dwarf_get_section_bytes(dbg,i,&elfScnIdx,&length,&err);
-    Elf_Scn* scn=elf_getscn(outelf,elfScnIdx);
+    Elf_Scn* scn=elf_getscn(patch->e,elfScnIdx);
     Elf_Data* data=elf_newdata(scn);
     ElfXX_Shdr* shdr=elfxx_getshdr(scn);
     shdr->sh_size=length;
@@ -835,14 +843,20 @@ void writePatch(char* oldSourceTree,char* newSourceTree,char* oldBinName,char* n
     memcpy(data->d_buf,buf,length);
   }
   
-  endPatchElf();
+  finalizeModifiedElf(patch);
   int res=dwarf_producer_finish(dbg,&err);
   if(DW_DLV_ERROR==res)
   {
     dwarfErrorHandler(err,NULL);
   }
-  logprintf(ELL_INFO_V1,ELS_ELFWRITE,"wrote elf file %s\n",patchOutName);
   endELF(oldBinary);
   endELF(newBinary);
+
+  if(elf_update (patch->e, ELF_C_WRITE) <0)
+  {
+    death("Failed to write out elf file: %s\n",elf_errmsg (-1));
+    exit(1);
+  }
+  return patch;
 }
 
